// this file handles outputting usage instructions,
// failures, etc. keeps logging in one place.
var tw = require('tty-wrap')
var decamelize = require('decamelize')
var EOL = require('os').EOL

module.exports = function (yargs, y18n) {
  var __ = y18n.__
  var self = {}

  // methods for ouputting/building failure message.
  var fails = []
  self.failFn = function (f) {
    fails.push(f)
  }

  var failMessage = null
  var showHelpOnFail = true
  self.showHelpOnFail = function (enabled, message) {
    if (typeof enabled === 'string') {
      message = enabled
      enabled = true
    } else if (typeof enabled === 'undefined') {
      enabled = true
    }
    failMessage = message
    showHelpOnFail = enabled
    return self
  }

  self.fail = function (msg) {
    if (fails.length) {
      fails.forEach(function (f) {
        f(msg)
      })
    } else {
      if (showHelpOnFail) yargs.showHelp('error')
      if (msg) console.error(msg)
      if (failMessage) {
        if (msg) console.error('')
        console.error(failMessage)
      }
      if (yargs.getExitProcess()) {
        process.exit(1)
      } else {
        throw new Error(msg)
      }
    }
  }

  // methods for ouputting/building help (usage) message.
  var usage
  self.usage = function (msg) {
    usage = msg
  }

  var examples = []
  self.example = function (cmd, description) {
    examples.push([cmd, description || ''])
  }

  var commands = []
  self.command = function (cmd, description) {
    commands.push([cmd, description || ''])
  }
  self.getCommands = function () {
    return commands
  }

  var descriptions = {}
  self.describe = function (key, desc) {
    if (typeof key === 'object') {
      Object.keys(key).forEach(function (k) {
        self.describe(k, key[k])
      })
    } else {
      descriptions[key] = desc
    }
  }
  self.getDescriptions = function () {
    return descriptions
  }

  var epilog
  self.epilog = function (msg) {
    epilog = msg
  }

  // var wrap = windowWidth()
  self.wrap = function (cols) {
    // wrap = cols
  }

  var deferY18nLookupPrefix = '__yargsString__:'
  self.deferY18nLookup = function (str) {
    return deferY18nLookupPrefix + str
  }

  self.help = function () {
    normalizeAliases()

    var demanded = yargs.getDemanded()
    var options = yargs.getOptions()
    var keys = Object.keys(
      Object.keys(descriptions)
      .concat(Object.keys(demanded))
      .concat(Object.keys(options.default))
      .reduce(function (acc, key) {
        if (key !== '_') acc[key] = true
        return acc
      }, {})
    )

    var output = []
    var head = function (h) { return tw.chalk.bold(__(h)) }
    var optionsTable = []

    // the usage string.
    if (usage) {
      output.push('', head('Usage:') + usage.replace(/\$0/g, yargs.$0), '')
    }

    // your application's commands, i.e., non-option
    // arguments populated in '_'.
    if (commands.length) {
      output.push(head('Commands:'), tw.table(commands, {left: 1, right: 10}, {colA: {color: 'green'}}), '')
    }

    // the options table.
    var aliasKeys = (Object.keys(options.alias) || [])
      .concat(Object.keys(yargs.parsed.newAliases) || [])

    keys = keys.filter(function (key) {
      return !yargs.parsed.newAliases[key] && aliasKeys.every(function (alias) {
        return (options.alias[alias] || []).indexOf(key) === -1
      })
    })

    var switches = keys.reduce(function (acc, key) {
      acc[key] = [ key ].concat(options.alias[key] || [])
      .map(function (sw) {
        if (/^\s{2,}/.test(sw)) return EOL + tw.chalk.bold(sw.trim()) + EOL;
        return (sw.length > 1 ? '--' : '-') + sw
      })
      .join(', ')

      return acc
    }, {})

    if (keys.length) {
      output.push(head('Options:'))

      keys.forEach(function (key) {
        var kswitch = switches[key]
        var desc = descriptions[key] || ''
        var type = null

        if (~desc.lastIndexOf(deferY18nLookupPrefix)) desc = __(desc.substring(deferY18nLookupPrefix.length))

        // if (~options.boolean.indexOf(key)) type = '[B]'
        if (~options.count.indexOf(key)) type = '[C]'
        if (~options.string.indexOf(key)) type = '[S]'
        if (~options.normalize.indexOf(key)) type = '[S]'
        if (~options.array.indexOf(key)) type = '[A]'

        var extra = defaultString(options.default[key], options.defaultDescription[key]) ||
          demanded[key] && '[ ' + __('required') + ' ]'
        extra = extra ? tw.chalk.gray(extra) : ''

        var choices = options.choices && options.choices[key] ? __('choices:') + ' ' +
          self.stringifiedValues(options.choices[key]) : null
        choices = choices ?  tw.chalk.gray.bold(' [ ' + choices + ' ]') : '';

        optionsTable.push([kswitch, type, desc + ' ' + extra + choices])
      })

      output.push(
        tw.table(optionsTable,
          {
            left: 1,
            right: 10,
            autoWrapOnCol: 'C'
          }, {
            colA: {color: 'green'},
            colB: {color: 'cyan'}
          }
        ),
        ''
      )
    }

    // describe some common use-cases for your application.
    if (examples.length) {
      output.push(head('Examples:'))

      examples.forEach(function (example) {
        example[0] = example[0].replace(/\$0/g, yargs.$0)
      })

      output.push(tw.table(examples, {left: 1}, {colB: {color: 'gray'}}), '')
    }

    // the usage string.
    if (epilog) {
      output.push(epilog.replace(/\$0/g, yargs.$0), '')
    }

    return output.join(EOL)
  }

  // make sure any options set for aliases,
  // are copied to the keys being aliased.
  function normalizeAliases () {
    var demanded = yargs.getDemanded()
    var options = yargs.getOptions()

    ;(Object.keys(options.alias) || []).forEach(function (key) {
      options.alias[key].forEach(function (alias) {
        // copy descriptions.
        if (descriptions[alias]) self.describe(key, descriptions[alias])
        // copy demanded.
        if (demanded[alias]) yargs.demand(key, demanded[alias].msg)

        // type messages.
        if (~options.boolean.indexOf(alias)) yargs.boolean(key)
        if (~options.count.indexOf(alias)) yargs.count(key)
        if (~options.string.indexOf(alias)) yargs.string(key)
        if (~options.normalize.indexOf(alias)) yargs.normalize(key)
        if (~options.array.indexOf(alias)) yargs.array(key)
      })
    })
  }

  self.showHelp = function (level) {
    level = level || 'error'
    console[level](self.help())
  }

  self.functionDescription = function (fn) {
    var description = fn.name ? decamelize(fn.name, '-') : __('generated-value')
    return ['(', description, ')'].join('')
  }

  self.stringifiedValues = function (values, separator) {
    var string = ''
    var sep = separator || ', '
    var array = [].concat(values)

    if (!values || !array.length) return string

    array.forEach(function (value) {
      if (string.length) string += sep
      string += JSON.stringify(value)
    })

    return string
  }

  // format the default-value-string displayed in
  // the right-hand column.
  function defaultString (value, defaultDescription) {
    var string = '[ ' + __('default:') + ' '

    if (value === undefined && !defaultDescription) return null

    if (defaultDescription) {
      string += defaultDescription
    } else {
      switch (typeof value) {
        case 'string':
          string += JSON.stringify(value)
          break
        case 'object':
          string += JSON.stringify(value)
          break
        default:
          string += value
      }
    }

    return string + ' ]'
  }

  // logic for displaying application version.
  var version = null
  self.version = function (ver, opt, msg) {
    version = ver
  }

  self.showVersion = function () {
    if (typeof version === 'function') console.log(version())
    else console.log(version)
  }

  return self
}
